@@ -13,6 +13,8 @@
 # limitations under the License.
 
 import io
+import logging
+from re import I
 import tarfile
 import time
 import typing
@@ -23,7 +25,9 @@ from .network import bind_floating_ip, get_free_floating_ip, get_network_id
 if typing.TYPE_CHECKING:
     from zunclient.v1.containers import Container
 
+DEFAULT_IMAGE_DRIVER = "docker"
 DEFAULT_NETWORK = "containernet1"
+LOG = logging.getLogger(__name__)
 
 __all__ = [
     "list_containers",
@@ -33,6 +37,7 @@ __all__ = [
     "destroy_container",
     "get_logs",
     "execute",
+    "upload",
     "wait_for_active",
 ]
 
@@ -40,14 +45,16 @@ __all__ = [
 def create_container(
     name: "str",
     image: "str" = None,
+    image_driver: "str" = DEFAULT_IMAGE_DRIVER,
     environment: "dict" = None,
     exposed_ports: "list[str]" = [],
     runtime: "str" = None,
-    host: "str" = None,
     nets: "list[dict]" = None,
     network_id: "str" = None,
     network_name: "str" = DEFAULT_NETWORK,
+    reservation_id: "str" = None,
     start: "bool" = True,
+    start_timeout: "int" = None,
     **kwargs,
 ) -> "Container":
     """Create a container instance.
@@ -56,6 +63,10 @@ def create_container(
         name (str): The name to give the container.
         image (str): The Docker image, with or without tag information. If no
             tag is provided, "latest" is assumed.
+        image_driver (str): The image storage driver to use to retrieve the
+            image. Defaults to "docker", meaning the image is assumed to be a
+            Docker registry repository. Specify "glance" to launch a snapshot
+            image by passing the Glance Image ID in the ``image`` argument.
         environment (dict): A set of environment variables to pass to the
             container.
         exposed_ports (list[str]): A list of ports to expose on the container.
@@ -87,20 +98,33 @@ def create_container(
             network_id = get_network_id(network_name)
         nets = [{"network": network_id}]
 
+    hints = {}
+    if reservation_id:
+        hints["reservation"] = reservation_id
+
+    # Note: ``host`` is not defined as an arg because there is some special
+    # handling of it in the Zun client; it is not sent if it is not on kwargs.
+    # If it is on kwargs it is expected to be non-None.
     container = zun().containers.create(
         name=name,
         image=image,
-        image_driver="docker",
+        image_driver=image_driver,
         nets=nets,
         exposed_ports={port_def: {} for port_def in (exposed_ports or [])},
         environment=environment,
         runtime=runtime,
-        host=host,
+        hints=hints,
         **kwargs,
     )
 
     if start:
-        container = _wait_for_status(container.uuid, "Created")
+        timeout = start_timeout or (60 * 30)
+        LOG.info(f"Waiting up to {timeout}s for container creation ...")
+        # Wait for a while, the image may need to download. 30 minutes is
+        # _quite_ a long time, but the user can interrupt or choose a smaller
+        # timeout.
+        container = _wait_for_status(container.uuid, "Created", timeout=timeout)
+        LOG.info("Starting container ...")
         zun().containers.start(container.uuid)
 
     return container
@@ -120,6 +144,8 @@ def get_container(container_ref: "str") -> "Container":
 
     Args:
         container_ref (str): The name or ID of the container.
+        tag (str): An optional version to tag the container image with. If not
+            defined, defaults to "latest".
 
     Returns:
         The container, if found.
@@ -128,8 +154,19 @@ def get_container(container_ref: "str") -> "Container":
 
 
 def snapshot_container(
-    container_ref: "str", repository: "str", tag: "str" = None
+    container_ref: "str", repository: "str", tag: "str" = "latest"
 ) -> "str":
+    """Create a snapshot of a running container.
+
+    This will store the container's file system in Glance as a new Image.
+    You can then specify the Image ID in container create requests.
+
+    Args:
+        container_ref (str): The name or ID of the container.
+        repository (str): The name to give the snapshot.
+        tag (str): An optional version tag to give the snapshot. Defaults to
+            "latest".
+    """
     return zun().containers.commit(container_ref, repository, tag=tag)["uuid"]
 
 
@@ -173,6 +210,13 @@ def execute(container_ref: "str", command: "str") -> "dict":
 
 
 def upload(container_ref: "str", source: "str", dest: "str") -> "dict":
+    """Upload a file or directory to a running container.
+
+    Args:
+        container_ref (str): The name or ID of the container.
+        source (str): The (local) path to the file or directory to upload.
+        dest (str): The (container) path to upload the file or directory to.
+    """
     fd = io.BytesIO()
     with tarfile.open(fileobj=fd, mode="w") as tar:
         tar.add(source, arcname=".")
